import httpx


class APIException(Exception):
    """Fortigate Base API Exception extends :py:class:`Exception`."""

    def __init__(self, response: httpx.Response) -> None:
        self.response = response
        super().__init__(self.message())

    def message(self):
        """Return formatted exception message with response code and detailed error description."""
        forti_error_msg: str = self.response.text
        if "cli_error" in forti_error_msg:
            forti_error_msg = self.response.json()["cli_error"]
        return repr(
            f"Response Code: {self.response.status_code} - "
            f"{forti_error_msg}"
        )


class BadRequestExeption(APIException):
    pass


class NotAuthorizedException(APIException):
    pass


class ForbiddenException(APIException):
    pass


class ResourceNotFoundException(APIException):
    pass


class MethodNotAllowedException(APIException):
    pass


class RequestEntityTooLargeException(APIException):
    pass


class FailedDependencyException(APIException):
    pass


class TooManyRequestsException(APIException):
    pass


class InternalErrorException(APIException):
    pass


class FortilibException(Exception):
    pass


class ObjectAlreadyExitsException(FortilibException):
    pass


class InterfaceMismatchException(FortilibException):
    pass


class AddressTypeMismatchException(FortilibException):
    pass



class FortigateObject:
    """Fortigate Base Object for all objects.

    :param object_data: raw dict of firewall object representation
    :param comment: comment of object
    """

    def __init__(self):
        self.object_data: dict = {}
        self.comment: str = ""

    @classmethod
    def from_dict(cls, object_data: dict):
        """Create fortigate class object direct with raw dict of firewall object representation.

        :param object_data: raw dict of firewall object representation
        """
        class_ = cls()
        class_.populate(object_data)

        return class_

    def populate(self, object_data: dict):
        """Parse raw dict data to fortigate base object.

        :param object_data: raw dict of firewall object representation
        """

        self.object_data = object_data
        self.comment = object_data.get("comment", self.comment)

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        .. warning:: Not implemented!

        """
        raise Exception("not implemented")

    def __repr__(self) -> str:
        return str(self.render())


class FortigateNamedObject(FortigateObject):
    """Fortigate Base Named Object extends :class:`fortilib.base.FortigateObject` with a name.

    .. inheritance-diagram:: fortilib.base.FortigateNamedObject
        :top-classes: fortilib.base.FortigateObject

    :param name: name of object
    """

    def __init__(self):
        super().__init__()

        self.name: str = ""

    def populate(self, object_data: dict):
        """Generate dict with all object arguments for fortigate api call."""
        super().populate(object_data)
        self.name = object_data.get("name", self.name)

    def __eq__(self, other):
        if isinstance(other, FortigateNamedObject):
            return self.name == other.name
        return False



from fortilib import FortilibSettings
from fortilib.base import FortigateNamedObject


class FortigateService(FortigateNamedObject):
    """Fortigate object for services.

    :ivar protocol: Name of internet protocol \
        ("TCP", "UDP", "TCP/UDP/SCTP", "ICMP", "ICMP6", "IP")
    """

    def __init__(self):
        super().__init__()

        self.protocol: str = ""
        self.visibility: str = "enable"

    def __eq__(self, other):
        if isinstance(other, FortigateService):
            return self.name == other.name and self.protocol == other.protocol

        return False

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.protocol = object_data.get("protocol", self.protocol)
        self.visibility = object_data.get("visibility", self.visibility)


class FortigateTCPUDPService(FortigateService):
    """Fortigate object for TCP/UDP services.

    :ivar protocol: Name of internet protocol (default: "TCP/UDP/SCTP")
    :ivar tcp_portrange: Port number for TCP (default: "")
    :ivar udp_portrange: Port number for UDP (default: "")
    """

    def __init__(self):
        super().__init__()

        self.protocol = str(FortilibSettings.tcp_udp_service_protocol)
        self.tcp_portrange: str = ""
        self.udp_portrange: str = ""

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.tcp_portrange = object_data.get(
            "tcp-portrange", self.tcp_portrange
        )
        self.udp_portrange = object_data.get(
            "udp-portrange", self.udp_portrange
        )

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "name": "test_service_tcp_udp",
                    "protocol": "TCP/UDP/SCTP",
                    "tcp-portrange": "80",
                    "udp-portrange": "",
                    "comment": "Test comment",
                }
        """
        return {
            "name": self.name,
            "protocol": self.protocol,
            "tcp-portrange": self.tcp_portrange,
            "udp-portrange": self.udp_portrange,
            "visibility": self.visibility,
            "comment": self.comment,
        }


class FortigateICMPService(FortigateService):
    """Fortigate object for ICMP services.

    :ivar protocol: Name of internet protocol (default: "ICMP")
    :ivar icmptype: ICMP type e.g. 8 (default: 0)
    :ivar icmpcode: ICMP code (default: "")
    """

    def __init__(self):
        super().__init__()

        self.protocol = "ICMP"
        self.icmptype: int = 0
        self.icmpcode: str = ""

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.icmptype = object_data.get("icmptype", self.icmptype)
        self.icmpcode = object_data.get("icmpcode", self.icmpcode)

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "name": "test_service_icmp",
                    "protocol": "ICMP",
                    "icmptype": 8,
                    "icmpcode": "",
                    "comment": "Test comment",
                }

        """
        return {
            "name": self.name,
            "protocol": self.protocol,
            "icmptype": self.icmptype,
            "icmpcode": self.icmpcode,
            "comment": self.comment,
        }


class FortigateICMP6Service(FortigateICMPService):
    """Fortigate object for ICMP6 services.

    :ivar protocol: Name of internet protocol (default: "ICMP6")
    """

    def __init__(self):
        super().__init__()

        self.protocol = "ICMP6"

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "name": "test_service_icmp6",
                    "protocol": "ICMP6",
                    "icmptype": "",
                    "comment": "Test comment",
                }

        """
        return {
            "name": self.name,
            "protocol": self.protocol,
            "icmptype": self.icmptype,
            "comment": self.comment,
        }


class FortigateIPService(FortigateService):
    """Fortigate object for IP services.

    :ivar protocol: Name of internet protocol (default: "IP")
    :ivar protocol_number: Protocol number (default: 0)
    """

    def __init__(self):
        super().__init__()

        self.protocol = "IP"
        self.protocol_number: int = 0

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.protocol_number = object_data.get(
            "protocol-number", self.protocol_number
        )

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "name": "test_service_ip",
                    "protocol": "IP",
                    "protocol-number": 47,
                    "comment": "Test comment",
                }

        """
        return {
            "name": self.name,
            "protocol": self.protocol,
            "protocol-number": self.protocol_number,
            "comment": self.comment,
        }


class FortigateProxyService(FortigateService):
    """Fortigate object for IP services.

    :ivar protocol: Name of internet protocol (default: "ALL")
    :ivar proxy: proxy flag (default: "enable")
    :ivar tcp_portrange: Port number for TCP (default: "")
    """

    def __init__(self):
        super().__init__()

        self.protocol = "ALL"
        self.proxy = "enable"
        self.tcp_portrange: str = ""

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.proxy = object_data.get("proxy", self.proxy)
        self.tcp_portrange = object_data.get(
            "tcp-portrange", self.tcp_portrange
        )

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "name": "test_service_proxy",
                    "protocol": "ALL",
                    "proxy": "enable",
                    "tcp-portrange": "443",
                    "comment": "Test comment",
                }

        """
        return {
            "name": self.name,
            "protocol": self.protocol,
            "proxy": self.proxy,
            "tcp-portrange": self.tcp_portrange,
            "comment": self.comment,
        }



from typing import List

from fortilib import get_by


class FortigateGroupMixin:
    def find_member(self, search_lists: List[List]):
        for member_raw in self.object_data.get("member", []):
            member = None
            for search_list in search_lists:
                member = get_by("name", member_raw["name"], search_list)
                if member is not None:
                    break

            if member is None:
                raise Exception(
                    f"group member '{member_raw['name']}' "
                    f"of group '{self.name}' not found"
                )

            self.member.append(member)



from typing import (
    Dict,
    List,
    Union,
)

from fortilib import (
    get_by,
    get_fortigate_member_array,
)
from fortilib.address import FortigateAddress
from fortilib.base import FortigateNamedObject
from fortilib.exceptions import (
    AddressTypeMismatchException,
    InterfaceMismatchException,
)
from fortilib.interface import FortigateInterface
from fortilib.proxyaddress import (
    FortigateProxyAddress,
    FortigateProxyAddressHostRegex,
    FortigateProxyAddressURL,
)
from fortilib.service import FortigateService
from fortilib.servicegroup import FortigateServiceGroup
from fortilib.vip import FortigateVIP
from fortilib.vipgroup import FortigateVIPGroup


class FortigateProxyPolicy(FortigateNamedObject):
    def __init__(self):
        super().__init__()

        self.policyid: int = 0
        self.action: str = "accept"
        self.status: str = "enable"

        self.srcintf: List[FortigateInterface] = []
        self.dstintf: List[FortigateInterface] = []

        self.srcaddr: List[FortigateAddress] = []
        self.dstaddr: List[Union[FortigateAddress, FortigateProxyAddress]] = []

        self.service: List[FortigateService] = []

        self.schedule: str = "always"
        self.proxy: str = "transparent-web"
        self.logtraffic: str = "all"
        self.utm_status: str = "enable"
        self.profile_type: str = "group"
        self.profile_group: str = "proxy_log"

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.policyid = object_data.get("policyid", self.policyid)
        self.action = object_data.get("action", self.action)
        self.status = object_data.get("status", self.status)

        self.schedule = object_data.get("schedule", self.schedule)
        self.proxy = object_data.get("proxy", self.proxy)
        self.logtraffic = object_data.get("logtraffic", self.logtraffic)
        self.utm_status = object_data.get("utm-status", self.utm_status)
        self.profile_type = object_data.get("profile-type", self.profile_type)
        self.profile_group = object_data.get(
            "profile-group", self.profile_group
        )
        self.comment = object_data.get("comments", self.comment)

    def find_interfaces(self, interfaces: List[FortigateInterface]):
        self.srcintf = self.find_interface_for(
            self.object_data.get("srcintf"),
            interfaces,
        )
        self.dstintf = self.find_interface_for(
            self.object_data.get("dstintf"),
            interfaces,
        )

    @staticmethod
    def find_interface_for(
        interface_arr: List[Dict],
        all_interfaces: List[FortigateInterface],
    ) -> List[FortigateInterface]:
        interfaces: List[FortigateInterface] = []
        for interface_dict in interface_arr:
            interface = get_by("name", interface_dict["name"], all_interfaces)
            if interface is None:
                raise Exception(
                    f"no interface found "
                    f"with name {interface_dict['name']}"
                )
            interfaces.append(interface)

        return interfaces

    def find_addresses(
        self,
        addresses: List[List[Union[FortigateAddress, FortigateProxyAddress]]],
    ):
        self.srcaddr = self.find_addresses_for(
            self.object_data.get("srcaddr"),
            addresses,
        )
        self.dstaddr = self.find_addresses_for(
            self.object_data.get("dstaddr"),
            addresses,
        )

    @staticmethod
    def find_addresses_for(
        address_arr: List[Dict],
        all_addresses: List[
            List[Union[FortigateAddress, FortigateProxyAddress]]
        ],
    ) -> List[Union[FortigateAddress, FortigateProxyAddress]]:
        addresses: List[Union[FortigateAddress, FortigateProxyAddress]] = []
        for address_dict in address_arr:
            address = None
            for search_list in all_addresses:
                address = get_by("name", address_dict["name"], search_list)
                if address is not None:
                    break

            if address is None:
                raise Exception(
                    f"address (or vip or group) "
                    f"{address_dict['name']} not found"
                )

            addresses.append(address)

        return addresses

    def find_services(
        self,
        all_services: List[
            Union[List[FortigateService], List[FortigateServiceGroup]]
        ],
    ):
        for service_dict in self.object_data.get("service", []):
            service = None
            for search_list in all_services:
                service = get_by("name", service_dict["name"], search_list)
                if service is not None:
                    break
            if service is None:
                raise Exception(
                    f"service with name {service_dict['name']} " f"not found"
                )

            self.service.append(service)

    def add_source_interface(self, interface: FortigateInterface):
        if self.check_addresses_interface(interface, self.srcaddr):
            self.srcintf.append(interface)

    def add_destination_interface(self, interface: FortigateInterface):
        if self.check_addresses_interface(interface, self.dstaddr):
            self.dstintf.append(interface)

    def check_addresses_interface(
        self, interface: FortigateInterface, addresses: List[FortigateAddress]
    ) -> bool:
        for address in addresses:
            if (
                not isinstance(address, FortigateProxyAddress)
                and not isinstance(address, FortigateProxyAddressHostRegex)
                and not isinstance(address, FortigateProxyAddressURL)
            ):
                if address.interface and address.interface != interface:
                    raise InterfaceMismatchException(
                        f"""
                        inteface mismatch between interface {interface.name} and
                        {address.name} ({address.interface.name}) in
                        policy {self.policyid} {self.name}
                    """
                    )
        return True

    def remove_source_interface(self, interface: FortigateInterface):
        self.srcintf.remove(interface)

    def remove_destination_interface(self, interface: FortigateInterface):
        self.dstintf.remove(interface)

    def add_source_address(self, address: FortigateAddress):
        if (
            not isinstance(address, FortigateProxyAddress)
            and not isinstance(address, FortigateProxyAddressHostRegex)
            and not isinstance(address, FortigateProxyAddressURL)
        ):
            if self.check_interface_address(
                address, self.srcintf
            ) and self.check_address_type(address, self.srcaddr):
                self.srcaddr.append(address)
        else:
            raise AddressTypeMismatchException(
                """
            proxy addresses can not be configured as a source address
            """
            )

    def add_destination_address(
        self, address: Union[FortigateAddress, FortigateProxyAddress]
    ):
        if (
            not isinstance(address, FortigateProxyAddress)
            and not isinstance(address, FortigateProxyAddressHostRegex)
            and not isinstance(address, FortigateProxyAddressURL)
        ):
            if self.check_interface_address(
                address, self.dstintf
            ) and self.check_address_type(address, self.dstaddr):
                self.dstaddr.append(address)
        else:
            self.dstaddr.append(address)

    def check_interface_address(
        self, address: FortigateAddress, interfaces: List[FortigateInterface]
    ) -> bool:
        # policy hat mehr als 1 interface, aber addressobject hat ein interface
        # fest gesetzt
        if address.interface and len(interfaces) > 1:
            raise InterfaceMismatchException(
                f"""
            only interface any is allowed on address {address.name} because
            policy {self.policyid} {self.name} has multiple interfaces
            """
            )

        # ein interface, weicht das von dem von der addresse ab?
        if address.interface and len(interfaces) == 1:
            for interface in interfaces:
                if interface != address.interface:
                    raise InterfaceMismatchException(
                        f"""
                    inteface mismatch between interface {interface.name} and
                    {address.name} ({address.interface.name}) in
                    policy {self.policyid} {self.name}
                    """
                    )
        return True

    def check_address_type(
        self,
        address: FortigateAddress,
        existing_addresses: List[FortigateAddress],
    ) -> bool:
        # bisher keine addressen? dann ists ok
        if len(existing_addresses) < 1:
            return True

        vip_tuple = (FortigateVIP, FortigateVIPGroup)

        if (
            isinstance(address, vip_tuple)
            and not isinstance(existing_addresses[0], vip_tuple)
        ) or (
            not isinstance(address, vip_tuple)
            and isinstance(existing_addresses[0], vip_tuple)
        ):
            raise AddressTypeMismatchException(
                f"addresses and address groups"
                f" can not be mixed with vip "
                f"and vip groups in policy "
                f"{self.policyid} {self.name}"
            )

        return True

    def remove_source_address(self, address: FortigateAddress):
        self.srcaddr.remove(address)

    def remove_destination_address(
        self, address: Union[FortigateAddress, FortigateProxyAddress]
    ):
        self.dstaddr.remove(address)

    def render(self) -> dict:
        return {
            "policyid": self.policyid,
            "name": self.name,
            "proxy": self.proxy,
            "srcintf": get_fortigate_member_array(self.srcintf),
            "dstintf": get_fortigate_member_array(self.dstintf),
            "srcaddr": get_fortigate_member_array(self.srcaddr),
            "dstaddr": get_fortigate_member_array(self.dstaddr),
            "service": get_fortigate_member_array(self.service),
            "action": self.action,
            "status": self.status,
            "schedule": self.schedule,
            "logtraffic": self.logtraffic,
            "utm-status": self.utm_status,
            "profile-type": self.profile_type,
            "profile-group": self.profile_group,
            "comments": self.comment,
        }


class FortiproxyPolicy(FortigateProxyPolicy):
    def __init__(self):
        super().__init__()

        self.proxy = None
        self.type = "transparent"

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.policyid = object_data.get("policyid", self.policyid)
        self.action = object_data.get("action", self.action)
        self.status = object_data.get("status", self.status)

        self.schedule = object_data.get("schedule", self.schedule)
        self.type = object_data.get("type", self.type)
        self.logtraffic = object_data.get("logtraffic", self.logtraffic)
        self.utm_status = object_data.get("utm-status", self.utm_status)
        self.profile_type = object_data.get("profile-type", self.profile_type)
        self.profile_group = object_data.get(
            "profile-group", self.profile_group
        )
        self.comment = object_data.get("comments", self.comment)

    def render(self) -> dict:
        return {
            "policyid": self.policyid,
            "name": self.name,
            "type": self.type,
            "srcintf": get_fortigate_member_array(self.srcintf),
            "dstintf": get_fortigate_member_array(self.dstintf),
            "srcaddr": get_fortigate_member_array(self.srcaddr),
            "dstaddr": get_fortigate_member_array(self.dstaddr),
            "service": get_fortigate_member_array(self.service),
            "action": self.action,
            "status": self.status,
            "schedule": self.schedule,
            "logtraffic": self.logtraffic,
            "utm-status": self.utm_status,
            "profile-type": self.profile_type,
            "profile-group": self.profile_group,
            "comments": self.comment,
        }



import ipaddress

from fortilib import remove_empty_dict_values
from fortilib.base import FortigateNamedObject


class FortigatePhase2KeylifeType:
    SECONDS = "seconds"
    KILOBYTES = "kbs"
    BOTH = "both"


class FortigatePhase2AddressType:
    IP = "ip"
    RANGE = "range"
    SUBNET = "subnet"


class FortigatePhase2Interface(FortigateNamedObject):
    """Fortigate object for phase2 interfaces.

    :ivar phase1_name: Name of phase1
    :ivar src_start_ip: Source address e.g. 10.0.0.10
    :ivar src_end_ip: Source address e.g. 10.0.0.20
    :ivar src_subnet: Source subnet e.g. 10.0.0.0/8
    :ivar src_addr_type: Source address type e.g. subnet
    :ivar dst_start_ip: Destination address e.g. 192.168.100.10
    :ivar dst_end_ip: Destination address e.g. 192.168.100.20
    :ivar dst_subnet: Destination subnet e.g. 192.168.100.0/24
    :ivar dst_addr_type: Destination address type e.g. subnet
    :ivar dhgrp: Diffie-Hellman group e.g. 20
    :ivar auto_negotiate: Auto-negotiate enable/disable (default: "disable")
    :ivar keepalive: Keepalive enable/disable (default: "disable")
    :ivar replay: Replay detection enable/disable (default: "enable")
    :ivar pfs: PFS feature enable/disable (default: "enable")
    :ivar keylife_seconds: Key lifetime in seconds for phase2 e.g. 14400
    :ivar keylife_kbs: Key lifetime in number of kilobytes for phase2 e.g. 5120
    :ivar keylife_type: Key lifetime type for phase2 e.g. seconds, kbs or both (default: "seconds")
    :ivar proposal: Encryption and authentication algorithms for proposal e.g. chacha20poly1305 aes256-sha512 aes256gcm
    """

    def __init__(self):
        super().__init__()

        self.phase1_name: str = ""
        self.dst_subnet: ipaddress.IPv4Network = None
        self.src_subnet: ipaddress.IPv4Network = None
        self.dst_start_ip: ipaddress.IPv4Address = None
        self.dst_end_ip: ipaddress.IPv4Address = None
        self.src_start_ip: ipaddress.IPv4Address = None
        self.src_end_ip: ipaddress.IPv4Address = None
        self.dst_addr_type: FortigatePhase2AddressType = (
            FortigatePhase2AddressType.SUBNET
        )
        self.src_addr_type: FortigatePhase2AddressType = (
            FortigatePhase2AddressType.SUBNET
        )
        self.dhgrp: str = ""
        self.auto_negotiate: str = "disable"
        self.keepalive: str = "disable"
        self.pfs: str = "enable"
        self.replay: str = "enable"
        self.keylife_seconds: int = None
        self.keylife_kbs: int = None
        self.keylife_type: FortigatePhase2KeylifeType = (
            FortigatePhase2KeylifeType.SECONDS
        )
        self.proposal: str = ""

    def __eq__(self, other):
        if isinstance(other, FortigatePhase2Interface):
            return (
                self.name == other.name
                and self.dst_subnet == other.dst_subnet
                and self.src_subnet == other.src_subnet
            )

        return False

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.phase1_name = object_data.get("phase1name", self.phase1_name)
        self.dst_addr_type = object_data.get(
            "dst-addr-type", FortigatePhase2AddressType.SUBNET
        )
        self.src_addr_type = object_data.get(
            "src-addr-type", FortigatePhase2AddressType.SUBNET
        )
        if self.dst_addr_type == "subnet":
            self.dst_subnet = ipaddress.ip_network(
                "{}/{}".format(
                    object_data.get("dst-subnet", "0.0.0.0/0").split()[0],
                    object_data.get("dst-subnet", "0.0.0.0/0").split()[1],
                )
            )

        if self.src_addr_type == FortigatePhase2AddressType.SUBNET:
            self.src_subnet = ipaddress.ip_network(
                "{}/{}".format(
                    object_data.get("src-subnet", "0.0.0.0/0").split()[0],
                    object_data.get("src-subnet", "0.0.0.0/0").split()[1],
                )
            )
        if (
            self.dst_addr_type == FortigatePhase2AddressType.IP
            or self.dst_addr_type == FortigatePhase2AddressType.RANGE
        ):
            self.dst_start_ip = ipaddress.ip_address(
                object_data.get("dst-start-ip", "0.0.0.0")
            )
            if self.dst_addr_type == FortigatePhase2AddressType.RANGE:
                self.dst_end_ip = ipaddress.ip_address(
                    object_data.get("dst-end-ip", "0.0.0.0")
                )
        if (
            self.src_addr_type == FortigatePhase2AddressType.IP
            or self.src_addr_type == FortigatePhase2AddressType.RANGE
        ):
            self.src_start_ip = ipaddress.ip_address(
                object_data.get("src-start-ip", "0.0.0.0")
            )
            if self.src_addr_type == FortigatePhase2AddressType.RANGE:
                self.src_end_ip = ipaddress.ip_address(
                    object_data.get("src-end-ip", "0.0.0.0")
                )

        self.dhgrp = object_data.get("dhgrp", self.dhgrp)
        self.auto_negotiate = object_data.get(
            "auto-negotiate", self.auto_negotiate
        )
        self.pfs = object_data.get("pfs", self.pfs)
        self.replay = object_data.get("replay", self.replay)
        self.keepalive = object_data.get("keepalive", self.keepalive)
        self.keylife_seconds = object_data.get(
            "keylifeseconds", self.keylife_seconds
        )
        self.keylife_kbs = object_data.get("keylifekbs", self.keylife_kbs)
        self.keylife_type = object_data.get("keylife-type", self.keylife_type)
        self.proposal = object_data.get("proposal", self.proposal)
        self.comment = object_data.get("comments", self.comment)

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "name": "vpn_phase2",
                    "phase1name": "vpn_phase1",
                    "dst-addr-type": "subnet"
                    "dst-subnet": "192.168.100.0/24",
                    "dst-start-ip": "192.168.100.1",
                    "dst-end-ip": "192.168.100.128",
                    "src-addr-type": "subnet"
                    "src-subnet": "10.0.0.0/8",
                    "src-start-ip": "10.0.0.1",
                    "src-end-ip": "10.0.0.128",
                    "dhgrp": "20",
                    "pfs": "enable",
                    "replay": "enable",
                    "keepalive": "disable",
                    "auto-negotiate": "enable",
                    "keylifeseconds": 14400,
                    "keylifekbs": 5120,
                    "keylife-type": "seconds",
                    "proposal": "chacha20poly1305 aes256-sha512 aes256gcm",
                    "comments": "",
                },
        """
        return remove_empty_dict_values(
            {
                "name": self.name,
                "phase1name": self.phase1_name,
                "dst-addr-type": self.dst_addr_type,
                "dst-subnet": str(self.dst_subnet) if self.dst_subnet else "",
                "dst-start-ip": (
                    str(self.dst_start_ip) if self.dst_start_ip else ""
                ),
                "dst-end-ip": str(self.dst_end_ip) if self.dst_end_ip else "",
                "src-addr-type": self.src_addr_type,
                "src-subnet": str(self.src_subnet) if self.src_subnet else "",
                "src-start-ip": (
                    str(self.src_start_ip) if self.src_start_ip else ""
                ),
                "src-end-ip": str(self.src_end_ip) if self.src_end_ip else "",
                "dhgrp": self.dhgrp,
                "pfs": self.pfs,
                "replay": self.replay,
                "keepalive": self.keepalive,
                "auto-negotiate": self.auto_negotiate,
                "keylifeseconds": self.keylife_seconds,
                "keylifekbs": self.keylife_kbs,
                "keylife-type": self.keylife_type,
                "proposal": self.proposal,
                "comments": self.comment,
            }
        )

    def __repr__(self):
        return f"{self.__class__.__name__} {self.name} Phase1 Name: {self.phase1_name}"



from enum import Enum
from typing import (
    Dict,
    List,
    Union,
)

from fortilib import (
    get_by,
    get_fortigate_member_array,
)
from fortilib.address import FortigateAddress
from fortilib.base import FortigateNamedObject
from fortilib.exceptions import (
    AddressTypeMismatchException,
    InterfaceMismatchException,
)
from fortilib.interface import FortigateInterface
from fortilib.ippool import FortigateIPPool
from fortilib.service import FortigateService
from fortilib.servicegroup import FortigateServiceGroup
from fortilib.vip import FortigateVIP
from fortilib.vipgroup import FortigateVIPGroup


class FortigatePolicyAction(Enum):
    ACCEPT = "accept"
    DENY = "deny"


class FortigatePolicyLogTraffic(Enum):
    DISABLE = "disable"
    UTM = "utm"
    ALL = "all"


class FortigatePolicy(FortigateNamedObject):
    def __init__(self):
        super().__init__()

        self.policyid: int = 0
        self.action: FortigatePolicyAction = FortigatePolicyAction.ACCEPT

        self.srcintf: List[FortigateInterface] = []
        self.dstintf: List[FortigateInterface] = []

        self.srcaddr: List[FortigateAddress] = []
        self.dstaddr: List[FortigateAddress] = []

        self.service: List[FortigateService] = []

        self.nat: str = "disable"
        self.ippool: str = "disable"
        self.poolname: List[FortigateIPPool] = []
        self.schedule: str = "always"
        self.logtraffic: FortigatePolicyLogTraffic = (
            FortigatePolicyLogTraffic.ALL
        )

    def __eq__(self, other):
        if isinstance(other, FortigatePolicy):
            return (
                self.policyid == other.policyid
                and self.name == other.name
                and self.srcintf == other.srcintf
                and self.dstintf == other.dstintf
                and self.srcaddr == other.srcaddr
                and self.dstaddr == other.dstaddr
                and self.service == other.service
            )

        return False

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.policyid = object_data.get("policyid", self.policyid)
        self.action = FortigatePolicyAction[
            object_data.get(
                "action", FortigatePolicyAction.ACCEPT.value
            ).upper()
        ]
        self.nat = object_data.get("nat", self.nat)
        self.ippool = object_data.get("ippool", self.ippool)

        self.schedule = object_data.get("schedule", self.schedule)

        self.logtraffic = FortigatePolicyLogTraffic[
            object_data.get(
                "logtraffic", FortigatePolicyLogTraffic.ALL.value
            ).upper()
        ]

        self.comment = object_data.get("comments", self.comment)

    def find_interfaces(self, interfaces: List[FortigateInterface]):
        self.srcintf = self.find_interface_for(
            self.object_data.get("srcintf"),
            interfaces,
        )
        self.dstintf = self.find_interface_for(
            self.object_data.get("dstintf"),
            interfaces,
        )

    @staticmethod
    def find_interface_for(
        interface_arr: List[Dict],
        all_interfaces: List[FortigateInterface],
    ) -> List[FortigateInterface]:
        interfaces: List[FortigateInterface] = []
        for interface_dict in interface_arr:
            interface = get_by("name", interface_dict["name"], all_interfaces)
            if interface is None:
                raise Exception(
                    f"no interface found "
                    f"with name {interface_dict['name']}"
                )
            interfaces.append(interface)

        return interfaces

    def find_addresses(self, addresses: List[List[FortigateAddress]]):
        self.srcaddr = self.find_addresses_for(
            self.object_data.get("srcaddr"),
            addresses,
        )
        self.dstaddr = self.find_addresses_for(
            self.object_data.get("dstaddr"),
            addresses,
        )

    @staticmethod
    def find_addresses_for(
        address_arr: List[Dict], all_addresses: List[List[FortigateAddress]]
    ) -> List[FortigateAddress]:
        addresses: List[FortigateAddress] = []
        for address_dict in address_arr:
            address = None
            for search_list in all_addresses:
                address = get_by("name", address_dict["name"], search_list)
                if address is not None:
                    break

            if address is None:
                raise Exception(
                    f"address (or vip or group) "
                    f"{address_dict['name']} not found"
                )

            addresses.append(address)

        return addresses

    def find_services(
        self,
        all_services: List[
            Union[List[FortigateService], List[FortigateServiceGroup]]
        ],
    ):
        for service_dict in self.object_data.get("service", []):
            service = None
            for search_list in all_services:
                service = get_by("name", service_dict["name"], search_list)
                if service is not None:
                    break
            if service is None:
                raise Exception(
                    f"service with name {service_dict['name']} not found"
                )

            self.service.append(service)

    def find_ippools(self, all_ippools: List[FortigateIPPool]):
        for ippool_dict in self.object_data.get("poolname", []):
            ippool = get_by("name", ippool_dict["name"], all_ippools)
            if ippool is None:
                raise Exception(
                    f"ippool with name {ippool_dict['name']} " f"not found"
                )

            self.poolname.append(ippool)

    def add_source_interface(self, interface: FortigateInterface):
        if self.check_addresses_interface(interface, self.srcaddr):
            self.srcintf.append(interface)

    def add_destination_interface(self, interface: FortigateInterface):
        if self.check_addresses_interface(interface, self.dstaddr):
            self.dstintf.append(interface)

    def check_addresses_interface(
        self, interface: FortigateInterface, addresses: List[FortigateAddress]
    ) -> bool:
        for address in addresses:
            if address.interface and address.interface != interface:
                raise InterfaceMismatchException(
                    f"""
                    inteface mismatch between interface {interface.name} and
                     {address.name} ({address.interface.name}) in
                     policy {self.policyid} {self.name}
                """
                )
        return True

    def remove_source_interface(self, interface: FortigateInterface):
        self.srcintf.remove(interface)

    def remove_destination_interface(self, interface: FortigateInterface):
        self.dstintf.remove(interface)

    def add_source_address(self, address: FortigateAddress):
        if self.check_interface_address(
            address, self.srcintf
        ) and self.check_address_type(address, self.srcaddr):
            self.srcaddr.append(address)

    def add_destination_address(self, address: FortigateAddress):
        if self.check_interface_address(
            address, self.dstintf
        ) and self.check_address_type(address, self.dstaddr):
            self.dstaddr.append(address)

    def check_interface_address(
        self, address: FortigateAddress, interfaces: List[FortigateInterface]
    ) -> bool:
        # policy hat mehr als 1 interface, aber addressobject hat ein interface
        # fest gesetzt
        if address.interface and len(interfaces) > 1:
            raise InterfaceMismatchException(
                f"""
            only interface any is allowed on address {address.name} because
            policy {self.policyid} {self.name} has multiple interfaces
            """
            )

        # ein interface, weicht das von dem von der addresse ab?
        if address.interface and len(interfaces) == 1:
            for interface in interfaces:
                if interface != address.interface:
                    raise InterfaceMismatchException(
                        f"""
                    inteface mismatch between interface {interface.name} and
                    {address.name} ({address.interface.name}) in
                    policy {self.policyid} {self.name}
                    """
                    )
        return True

    def check_address_type(
        self,
        address: FortigateAddress,
        existing_addresses: List[FortigateAddress],
    ) -> bool:
        # bisher keine addressen? dann ists ok
        if len(existing_addresses) < 1:
            return True

        vip_tuple = (FortigateVIP, FortigateVIPGroup)

        if (
            isinstance(address, vip_tuple)
            and not isinstance(existing_addresses[0], vip_tuple)
        ) or (
            not isinstance(address, vip_tuple)
            and isinstance(existing_addresses[0], vip_tuple)
        ):
            raise AddressTypeMismatchException(
                f"addresses and address groups"
                f" can not be mixed with vip "
                f"and vip groups in policy "
                f"{self.policyid} {self.name}"
            )

        return True

    def remove_source_address(self, address: FortigateAddress):
        self.srcaddr.remove(address)

    def remove_destination_address(self, address: FortigateAddress):
        self.dstaddr.remove(address)

    def render(self) -> dict:
        return {
            "policyid": self.policyid,
            "name": self.name,
            "action": self.action.value,
            "nat": self.nat,
            "ippool": self.ippool,
            "srcintf": get_fortigate_member_array(self.srcintf),
            "dstintf": get_fortigate_member_array(self.dstintf),
            "srcaddr": get_fortigate_member_array(self.srcaddr),
            "dstaddr": get_fortigate_member_array(self.dstaddr),
            "service": get_fortigate_member_array(self.service),
            "poolname": get_fortigate_member_array(self.poolname),
            "schedule": self.schedule,
            "logtraffic": self.logtraffic.value,
            "comments": self.comment,
        }



import ipaddress

from fortilib.base import FortigateObject
from fortilib.interface import FortigateInterface
from fortilib.mixins.interface import FortigateInterfaceMixin


class FortigateStaticRoute(FortigateObject, FortigateInterfaceMixin):
    """Fortigate object for static routes.

    :ivar interface_attribute: (default: "device")
    :ivar status: Status - en-/disable static routes (default: "enable")
    :ivar seq_num: Sequence number - is unique for every route like an id (default: 0)
    :ivar dst: Destination - routing subnet
    :ivar gateway: Gateway
    :ivar distance: (default: 10)
    :ivar weight: (default: 0)
    :ivar priority: (default: 1)
    :ivar interface: Interface the static route belongs to
    """

    interface_attribute = "device"

    def __init__(self):
        super().__init__()

        self.status = "enable"
        self.seq_num: int = 0
        self.dst: ipaddress.IPv4Network = None
        self.gateway: ipaddress.IPv4Address = None
        self.distance: int = 10
        self.weight: int = 0
        self.priority: int = 1
        self.interface: FortigateInterface = None

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.status = object_data.get("status", self.status)
        self.seq_num = object_data.get("seq-num", self.seq_num)
        self.dst = ipaddress.ip_network(
            "{}/{}".format(
                object_data.get("dst", "0.0.0.0/0").split()[0],
                object_data.get("dst", "0.0.0.0/0").split()[1],
            )
        )
        self.gateway = ipaddress.IPv4Address(
            object_data.get("gateway", self.gateway)
        )

        self.distance = object_data.get("distance", self.distance)
        self.weight = object_data.get("weight", self.weight)
        self.priority = object_data.get("priority", self.priority)

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "status": "enable",
                    "seq-num": 0,
                    "dst": "10.0.0.0 255.0.0.0",
                    "gateway": "2.235.23.16",
                    "distance": 10,
                    "weight": 0,
                    "priority": 1,
                    "device": "port4",
                    "comment": "Test comment",
                }
        """
        return {
            "status": self.status,
            "seq-num": self.seq_num,
            "dst": f"{self.dst.network_address} {self.dst.netmask}",
            "gateway": str(self.gateway),
            "distance": self.distance,
            "weight": self.weight,
            "priority": self.priority,
            "device": self.interface.name if self.interface else "",
            "comment": self.comment,
        }

    def is_enabled(self) -> bool:
        return self.status == "enable"

    def __eq__(self, other):
        if isinstance(other, FortigateStaticRoute):
            return self.render() == other.render()
        return False



import ipaddress

from fortilib.address import FortigateAddress
from fortilib.mixins.interface import FortigateInterfaceMixin


class FortigateVIP(FortigateAddress, FortigateInterfaceMixin):
    """Fortigate object for VIPs.

    :ivar interface_attribute: Overwrite variable of :class:`fortilib.mixin.interface.FortigateInterfaceMixin` (default: "extintf")
    :ivar extip: ip address (usually external/public) of starting range e.g. 172.29.210.128
    :ivar extip_end: ip address (usually external/public) of ending range e.g. 172.29.210.143
    :ivar mappedip: ip address (usually internal/private) of starting range e.g. 10.10.210.128
    :ivar mappedip_end: ip address (usually internal/private) of ending range e.g. 10.10.210.143
    :ivar extport: port (usually external/public) e.g. 8080
    :ivar mappedport: port (usually internal/private) e.g. 80
    :ivar protocol: "tcp", "udp" or "icmp"
    :ivar portforward: "enable" or "disable"
    """

    interface_attribute = "extintf"

    def __init__(self):
        super().__init__()

        self.extip: ipaddress.IPv4Address = None
        self.extip_end: ipaddress.IPv4Address = None
        self.mappedip: ipaddress.IPv4Address = None
        self.mappedip_end: ipaddress.IPv4Address = None
        self.extport: str = "0-65535"
        self.mappedport: str = "0-65535"
        self.protocol: str = ""
        self.portforward: str = "disable"
        self.type: str = ""

    def populate(self, object_data: dict):
        """Parse raw dict data to vip object.

        :param object_data: raw dict of firewall object representation
        """

        super().populate(object_data)

        extip_split = object_data.get("extip").split("-")
        mappedip_split = object_data.get("mappedip")[0]["range"].split("-")

        self.extip = ipaddress.ip_address(
            extip_split[0],
        )
        self.mappedip = ipaddress.ip_address(
            mappedip_split[0],
        )

        if len(extip_split) == 2:
            self.extip_end = ipaddress.ip_address(
                extip_split[1],
            )
            self.mappedip_end = ipaddress.ip_address(
                mappedip_split[1],
            )

        self.extport = object_data.get("extport", self.extport)
        self.mappedport = object_data.get("mappedport", self.mappedport)
        self.protocol = object_data.get("protocol", self.protocol)
        self.portforward = object_data.get("portforward", self.portforward)
        self.type = object_data.get("type", self.type)

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "name": "test_vip",
                    "extip": "172.29.210.128-172.29.210.143"
                    "mappedip": [
                        {
                            "range": "10.10.210.128-10.10.210.143",
                        }
                    ],
                    "interface": "any",
                    "comments": "Test comment",
                }

        """
        if self.extip_end is None and self.extip_end is None:
            return {
                "name": self.name,
                "extip": f"{self.extip}",
                "mappedip": [
                    {
                        "range": f"{self.mappedip}",
                    }
                ],
                "comment": self.comment,
                "extintf": self.interface.name if self.interface else "any",
                "portforward": self.portforward,
                "protocol": self.protocol,
                "extport": self.extport,
                "mappedport": self.mappedport,
                "color": self.color,
            }
        else:
            return {
                "name": self.name,
                "extip": f"{self.extip}-{self.extip_end}",
                "mappedip": [
                    {
                        "range": f"{self.mappedip}-{self.mappedip_end}",
                    }
                ],
                "comment": self.comment,
                "extintf": self.interface.name if self.interface else "any",
                "portforward": self.portforward,
                "protocol": self.protocol,
                "extport": self.extport,
                "mappedport": self.mappedport,
                "color": self.color,
            }

    def __repr__(self):
        return f"FortigateVIP {self.name}"

    def __eq__(self, other):
        if isinstance(other, FortigateVIP):
            return (
                self.name,
                self.extip,
                self.extip_end,
                self.mappedip,
                self.mappedip_end,
                self.extport,
                self.mappedport,
            ) == (
                other.name,
                other.extip,
                other.extip_end,
                other.mappedip,
                other.mappedip_end,
                other.extport,
                other.mappedport,
            )
        return False

    def __hash__(self):
        return hash(
            (
                self.name,
                self.extip,
                self.extip_end,
                self.mappedip,
                self.mappedip_end,
                self.extport,
                self.mappedport,
            )
        )



import ipaddress

from fortilib.base import FortigateNamedObject
from fortilib.interface import FortigateInterface
from fortilib.mixins.interface import FortigateInterfaceMixin


class FortigateAddress(FortigateNamedObject, FortigateInterfaceMixin):
    interface_attribute = "associated-interface"

    """Fortigate object for addresses.

    :ivar interface: Interface the address is connected to
    """

    def __init__(self):
        super().__init__()

        self.interface: FortigateInterface = None
        self.color: int = 0

    # TODO mabye not need because super-element FortigateNamedObject implements it already -> delete?!
    def populate(self, object_data: dict):
        super().populate(object_data)

        self.color = object_data.get("color", self.color)


class FortigateIpMask(FortigateAddress):
    """Fortigate object for ip mask \
        extends :class:`fortilib.address.FortigateAddress` with a subnet.

    :ivar subnet: Adress subnet e.g. "10.10.10.1/255.255.255.255"
    """

    def __init__(self):
        super().__init__()
        self.subnet: ipaddress.IPv4Network = None

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.subnet = ipaddress.ip_network(
            "{}/{}".format(
                object_data.get("subnet", "0.0.0.0/0").split()[0],
                object_data.get("subnet", "0.0.0.0/0").split()[1],
            )
        )

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "name": "Test_address",
                    "type": "ipmask",
                    "subnet": "10.10.10.1 255.255.255.255",
                    "associated-interface": "port4",
                    "comments": "Test comment",
                }
        """
        return {
            "name": self.name,
            "type": "ipmask",
            "subnet": f"{self.subnet.network_address} {self.subnet.netmask}",
            "comment": self.comment,
            "associated-interface": (
                self.interface.name if self.interface else ""
            ),
            "color": self.color,
        }


class FortigateIpRange(FortigateAddress):
    """Fortigate object for ip ranges \
        extends :class:`fortilib.address.FortigateAddress` with a ip range.

    :ivar ip_start: Start of ip range
    :ivar ip_end: End of ip range
    """

    def __init__(self):
        super().__init__()
        self.ip_start: ipaddress.IPv4Address = None
        self.ip_end: ipaddress.IPv4Address = None

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.ip_start = ipaddress.ip_address(
            object_data.get("start-ip", self.ip_start)
        )
        self.ip_end = ipaddress.ip_address(
            object_data.get("end-ip", self.ip_end)
        )

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "name": "Test_ip_range",
                    "type": "iprange",
                    "start-ip": "10.10.10.1",
                    "end-ip": "10.10.10.15",
                    "associated-interface": "port4",
                    "comments": "Test comment",
                }
        """
        return {
            "name": self.name,
            "type": "iprange",
            "start-ip": str(self.ip_start),
            "end-ip": str(self.ip_end),
            "comment": self.comment,
            "associated-interface": (
                self.interface.name if self.interface else ""
            ),
            "color": self.color,
        }


class FortigateFQDN(FortigateAddress):
    """Fortigate object for ip ranges \
        extends :class:`fortilib.address.FortigateAddress` with a ip range.

    :ivar fqdn: \
        Fully qualified domain name (FQDN) - use domain name e.g. "kernel.org"
    """

    def __init__(self):
        super().__init__()
        self.fqdn: str = ""

    def populate(self, object_data: dict):
        super().populate(object_data)

        self.fqdn = object_data.get("fqdn", self.fqdn)

    def render(self) -> dict:
        """Generate dict with all object arguments for fortigate api call.

        :example:
            .. code-block:: json

                {
                    "name": "Test_fqdn",
                    "type": "fqdn",
                    "fqdn": "kernel.org",
                    "associated-interface": "port4",
                    "comments": "Test comment",
                }
        """
        return {
            "name": self.name,
            "type": "fqdn",
            "fqdn": self.fqdn,
            "comment": self.comment,
            "associated-interface": (
                self.interface.name if self.interface else ""
            ),
            "color": self.color,
        }



from typing import List

from fortilib.firewall import FortigateFirewall
from fortilib.fortigateapi import (
    FortiGateApiPolicyDirection,
    FortigateFirewallApi,
)
from fortilib.proxypolicy import FortiproxyPolicy


class Fortiproxy(FortigateFirewall):
    """Fortigate Firewall object.
    Representation of all firewall objects.

    :param fortigateapi: Fortigate API object (:class:`fortilib.fortigateapi.FortigateFirewallApi`) for API Querys
    :ivar name: Name of fortigate firewall
    :ivar fortigate: Fortigate API object (:class:`fortilib.fortigateapi.FortigateFirewallApi`) for API Querys
    :ivar interfaces: List of :class:`fortilib.interface.FortigateInterface` (default: [])
    :ivar static_routes: List of :class:`fortilib.routes.FortigateStaticRoute` (default: [])
    :ivar addresses: List of :class:`fortilib.address.FortigateAddress` (default: [])
    :ivar address_groups: List of :class:`fortilib.addressgroup.FortigateAddressGroup` (default: [])
    :ivar vips: List of :class:`fortilib.vip.FortigateVIP` (default: [])
    :ivar vip_groups: List of :class:`fortilib.vipgroup.FortigateVIPGroup` (default: [])
    :ivar services: List of :class:`fortilib.service.FortigateService` (default: [])
    :ivar service_groups: List of :class:`fortilib.servicegroup.FortigateServiceGroup` (default: [])
    :ivar ippools: List of :class:`fortilib.ippool.FortigateIPPool` (default: [])
    :ivar policies: List of :class:`fortilib.policy.FortigatePolicy` (default: [])
    :ivar proxy_addresses: List of :class:`fortilib.proxyaddresses.FortigateProxyAddress` (default: [])
    :ivar proxy_address_groups: List of :class:`fortilib.proxyaddressgroup.FortigateProxyAddressGroup` (default: [])
    :ivar proxy_policies: None
    :ivar all_addresses: List of :class:`fortilib.proxyaddresses.FortigateProxyAddress` and :class:`fortilib.address.FortigateAddress` (default: [])
    :ivar phase1_interfaces: None
    :ivar phase2_interfaces: None
    """

    def __init__(self, name: str, fortigateapi: FortigateFirewallApi):
        super().__init__(name, fortigateapi)

        self.policies: List[FortiproxyPolicy] = []

        self.proxy_policies = None
        self.phase1_interfaces = None
        self.phase2_interfaces = None

    def get_all_objects(self):
        """Query Fortiproxy API for all firewall objects of the following list.

        - Interfaces
        - Static Routes
        - Addresses
        - Address Groups
        - VIPs
        - VIP Groups
        - Services
        - Service Groups
        - IPPools
        - Policies
        - Proxy Addresses
        - Proxy Address Groups
        - Addresses and Proxy Addresses
        """
        self.interfaces = self.get_interfaces()
        self.static_routes = self.get_static_routes()
        self.addresses = self.get_addresses()
        self.address_groups = self.get_address_groups()
        self.resolve_address_groups()
        self.proxy_addresses = self.get_proxy_addresses()
        self.proxy_address_groups = self.get_proxy_address_groups()
        self.resolve_proxy_address_groups()
        self.vips = self.get_vips()
        self.vip_groups = self.get_vip_groups()
        self.resolve_vip_groups()
        self.services = self.get_services()
        self.service_groups = self.get_service_groups()
        self.resolve_service_groups()
        self.ippools = self.get_ippools()
        self.policies = self.get_policies()
        self.all_addresses = self.get_all_addresses()

    def get_policies(self) -> List[FortiproxyPolicy]:
        """Query Fortiproxy API for policies
        :obj:`fortilib.policy.FortigatePolicy` and create list.
        """
        policies: List[FortiproxyPolicy] = []
        for raw in self.fortigate.get_firewall_policies():
            policy = FortiproxyPolicy.from_dict(raw)

            policy.find_interfaces(self.interfaces)
            policy.find_addresses(
                [
                    self.addresses,
                    self.address_groups,
                    self.vips,
                    self.vip_groups,
                    self.proxy_addresses,
                    self.proxy_address_groups,
                ]
            )
            policy.find_services([self.services, self.service_groups])

            policies.append(policy)

        return policies

    def create_firewall_policy(self, policy: FortiproxyPolicy):
        """Create policy on fortigate with given :obj:`fortilib.proxypolicy.FortiproxyPolicy`."""
        status = self.fortigate.create_firewall_policy(
            int(policy.policyid), policy.render()
        )

        policy.policyid = status["mkey"]

        self.policies.append(policy)
        return status

    def update_firewall_policy(self, policy: FortiproxyPolicy):
        """Update policy on fortiproxy with given :obj:`fortilib.proxypolicy.FortiproxyPolicy`."""
        return self.fortigate.update_firewall_policy(
            policy.policyid, policy.render()
        )

    def delete_firewall_policy(self, policy: FortiproxyPolicy):
        """Delete policy on fortiproxy with given :obj:`fortilib.proxypolicy.FortiproxyPolicy`."""
        status = self.fortigate.delete_firewall_policy(policy.policyid)
        self.policies.remove(policy)
        return status

    def move_firewall_policy(
        self,
        policy: FortiproxyPolicy,
        move_direction: FortiGateApiPolicyDirection,
        move_identifier_policy: FortiproxyPolicy,
    ):
        """move policy on fortiproxy to another position with given :obj:`fortilib.proxypolicy.FortiproxyPolicy`, :obj:`fortilib.fortigateapi.FortiGateApiPolicyDirection` and :obj:`fortilib.policy.FortigatePolicy`."""

        return self.fortigate.move_firewall_policy(
            policy.policyid,
            move_direction,
            move_identifier_policy.policyid,
        )

    def get_proxy_policies(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.get_proxy_policies.__name__}",
            obj=Fortiproxy,
            name=f"{self.get_proxy_policies.__name__}",
        )

    def update_firewall_proxy_policy(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.update_firewall_proxy_policy.__name__}",
            obj=Fortiproxy,
            name=f"{self.update_firewall_proxy_policy.__name__}",
        )

    def delete_firewall_proxy_policy(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.delete_firewall_proxy_policy.__name__}",
            obj=Fortiproxy,
            name=f"{self.delete_firewall_proxy_policy.__name__}",
        )

    def move_firewall_proxy_policy(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.move_firewall_proxy_policy.__name__}",
            obj=Fortiproxy,
            name=f"{self.move_firewall_proxy_policy.__name__}",
        )

    def get_phase1_interfaces(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.get_phase1_interfaces.__name__}",
            obj=Fortiproxy,
            name=f"{self.get_phase1_interfaces.__name__}",
        )

    def create_firewall_phase1_interface(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.create_firewall_phase1_interface.__name__}",
            obj=Fortiproxy,
            name=f"{self.create_firewall_phase1_interface.__name__}",
        )

    def update_firewall_phase1_interface(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.update_firewall_phase1_interface.__name__}",
            obj=Fortiproxy,
            name=f"{self.update_firewall_phase1_interface.__name__}",
        )

    def delete_firewall_phase1_interface(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.delete_firewall_phase1_interface.__name__}",
            obj=Fortiproxy,
            name=f"{self.delete_firewall_phase1_interface.__name__}",
        )

    def get_phase2_interfaces(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.get_phase2_interfaces.__name__}",
            obj=Fortiproxy,
            name=f"{self.get_phase2_interfaces.__name__}",
        )

    def create_firewall_phase2_interface(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.create_firewall_phase2_interface.__name__}",
            obj=Fortiproxy,
            name=f"{self.create_firewall_phase2_interface.__name__}",
        )

    def update_firewall_phase2_interface(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.update_firewall_phase2_interface.__name__}",
            obj=Fortiproxy,
            name=f"{self.update_firewall_phase2_interface.__name__}",
        )

    def delete_firewall_phase2_interface(self):
        raise AttributeError(
            f"{Fortiproxy} has no Method {self.delete_firewall_phase2_interface.__name__}",
            obj=Fortiproxy,
            name=f"{self.delete_firewall_phase2_interface.__name__}",
        )



import urllib.parse
from enum import Enum
from typing import (
    Dict,
    List,
    Optional,
    Union,
)

import httpx

from fortilib.exceptions import (
    APIException,
    BadRequestExeption,
    FailedDependencyException,
    ForbiddenException,
    InternalErrorException,
    MethodNotAllowedException,
    NotAuthorizedException,
    ObjectAlreadyExitsException,
    RequestEntityTooLargeException,
    ResourceNotFoundException,
    TooManyRequestsException,
)


class FortiGateApiPolicyDirection(Enum):
    BEFORE = "before"
    AFTER = "after"


class FortigateFirewallApi:
    """Fortigate Firewall API object.

    :ivar ipaddr: IP address of fortigate firewall
    :ivar username: Login User
    :ivar password: Login Password
    :ivar vdom: Set default vdom that is used for all api calls.
    :ivar timeout: Set timeout for api calls. (default: 10)
    """

    def __init__(
        self,
        ipaddr: str,
        vdom: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        timeout: int = 10,
        read_only: bool = False,
        access_token: Optional[str] = None,
        port: int = 443,
    ):
        self.ipaddr = ipaddr
        self.username = username
        self.password = password
        self.access_token = access_token
        self.vdom = vdom
        self.timeout = timeout
        self.read_only: bool = read_only
        self.port = port

        self.fortigate: FortiGateApi

    def login(self):
        """Login into Fortigate API via :meth:`fortilib.fortigateapi.FortigateFirewallApi.login`."""
        self.fortigate = FortiGateApi(
            ipaddr=self.ipaddr,
            username=self.username,
            password=self.password,
            access_token=self.access_token,
            vdom=self.vdom,
            timeout=self.timeout,
            read_only=self.read_only,
            port=str(self.port),
        )
        self.fortigate.login()

    def logout(self):
        """Logout from Fortigate API via :meth:`fortilib.fortigateapi.FortigateFirewallApi.logout`."""
        self.fortigate.logout()

    def get_firewall_address(self):
        """Get addresses via Fortigate API"""
        return self.fortigate.get_firewall_address()

    def create_firewall_address(self, address_name: str, address_object: Dict):
        """Create address via Fortigate API

        :param address_name: Name of address on firewall
        :param address_object: Dict representation of address
        """
        return self.fortigate.create_firewall_address(
            address_name, address_object
        )

    def update_firewall_address(self, address_name: str, address_object: Dict):
        """Update address via Fortigate API

        :param address_name: Name of address on firewall
        :param address_object: Dict representation of address
        """
        return self.fortigate.update_firewall_address(
            address_name, address_object
        )

    def delete_firewall_address(self, address_name: str):
        """Delete address via Fortigate API

        :param address_name: Name of address on firewall
        """
        return self.fortigate.delete_firewall_address(address_name)

    def get_firewall_address_group(self):
        """Get address groups via Fortigate API"""
        return self.fortigate.get_address_group()

    def create_firewall_address_group(
        self, group_name: str, group_object: Dict
    ):
        """Create address group via Fortigate API

        :param group_name: Name of address group on firewall
        :param group_object: Dict representation of address group
        """
        return self.fortigate.create_address_group(
            group_name,
            group_object,
        )

    def update_firewall_address_group(
        self, group_name: str, group_object: Dict
    ):
        """Update address group via Fortigate API

        :param group_name: Name of address group on firewall
        :param group_object: Dict representation of address group
        """
        return self.fortigate.update_address_group(
            group_name,
            group_object,
        )

    def delete_firewall_address_group(self, group_name: str):
        """Delete address group via Fortigate API

        :param group_name: Name of address group on firewall
        """
        return self.fortigate.delete_address_group(group_name)

    def get_firewall_vip(self):
        """Get vips via Fortigate API"""
        return self.fortigate.get_firewall_vip()

    def create_firewall_vip(self, vip_name: str, vip_object: Dict):
        """Create vip via Fortigate API

        :param vip_name: Name of vip on firewall
        :param vip_object: Dict representation of vip
        """
        return self.fortigate.create_firewall_vip(vip_name, vip_object)

    def update_firewall_vip(self, vip_name: str, vip_object: Dict):
        """Update vip via Fortigate API

        :param vip_name: Name of vip on firewall
        :param vip_object: Dict representation of vip
        """
        return self.fortigate.update_firewall_vip(vip_name, vip_object)

    def delete_firewall_vip(self, vip_name: str):
        """Delete vip via Fortigate API

        :param vip_name: Name of vip on firewall
        """
        return self.fortigate.delete_firewall_vip(vip_name)

    def get_firewall_vip_group(self):
        """Get vip groups via Fortigate API"""
        return self.fortigate.get_firewall_vip_group()

    def create_firewall_vip_group(
        self, vip_group_name: str, vip_group_object: Dict
    ):
        """Create vip group via Fortigate API

        :param vip_group_name: Name of vip group on firewall
        :param vip_group_object: Dict representation of vip group
        """
        return self.fortigate.create_firewall_vip_group(
            vip_group_name, vip_group_object
        )

    def update_firewall_vip_group(
        self, vip_group_name: str, vip_group_object: Dict
    ):
        """Update vip group via Fortigate API

        :param vip_group_name: Name of vip group on firewall
        :param vip_group_object: Dict representation of vip group
        """
        return self.fortigate.update_firewall_vip_group(
            vip_group_name, vip_group_object
        )

    def delete_firewall_vip_group(self, vip_group_name: str):
        """Delete vip group via Fortigate API

        :param vip_group_name: Name of vip group on firewall
        :param vip_group_object: Dict representation of vip group
        """
        return self.fortigate.delete_firewall_vip_group(vip_group_name)

    def get_firewall_interface(self):
        """Get interfaces via Fortigate API"""
        return self.fortigate.get_firewall_interface()

    def get_firewall_route_static(self):
        """Get static routes via Fortigate API"""
        return self.fortigate.get_firewall_route_static()

    def create_firewall_route_static(self, seq_num: str, route_object: Dict):
        """Create static route via Fortigate API

        :param seq_num: Sequence number of static route on firewall
        :param route_object: Dict representation of static route
        """
        return self.fortigate.create_firewall_route_static(
            seq_num, route_object
        )

    def update_firewall_route_static(self, seq_num: str, route_object: Dict):
        """Update static route via Fortigate API

        :param seq_num: Sequence number of static route on firewall
        :param route_object: Dict representation of static route
        """
        return self.fortigate.update_firewall_route_static(
            seq_num, route_object
        )

    def delete_firewall_route_static(self, seq_num: str):
        """Delete static route via Fortigate API

        :param seq_num: Sequence number of static route on firewall
        """
        return self.fortigate.delete_firewall_route_static(seq_num)

    def get_firewall_service(self):
        """Get services via Fortigate API"""
        return self.fortigate.get_firewall_service()

    def create_firewall_service(self, service_name: str, service_object: Dict):
        """Create service via Fortigate API

        :param service_name: Name of service on firewall
        :param service_object: Dict representation of service
        """
        return self.fortigate.create_firewall_service(
            service_name, service_object
        )

    def update_firewall_service(self, service_name: str, service_object: Dict):
        """Update service via Fortigate API

        :param service_name: Name of service on firewall
        :param service_object: Dict representation of service
        """
        return self.fortigate.update_firewall_service(
            service_name, service_object
        )

    def delete_firewall_service(self, service_name: str):
        """Delete service via Fortigate API

        :param service_name: Name of service on firewall
        """
        return self.fortigate.delete_firewall_service(service_name)

    def create_firewall_service_group(
        self, group_name: str, group_object: Dict
    ):
        """Create service group via Fortigate API

        :param group_name: Name of service group on firewall
        :param group_object: Dict representation of service group
        """
        return self.fortigate.create_service_group(group_name, group_object)

    def update_firewall_service_group(
        self, group_name: str, group_object: Dict
    ):
        """Update service group via Fortigate API

        :param group_name: Name of service group on firewall
        :param group_object: Dict representation of service group
        """
        return self.fortigate.update_service_group(group_name, group_object)

    def delete_firewall_service_group(self, group_name: str):
        """Delete service group via Fortigate API

        :param group_name: Name of service group on firewall
        """
        return self.fortigate.delete_service_group(group_name)

    def get_firewall_service_groups(self):
        """Get service groups via Fortigate API"""
        return self.fortigate.get_service_group()

    def get_firewall_ippool(self):
        """Get ip pools via Fortigate API"""
        return self.fortigate.get_firewall_ippool()

    def create_firewall_ippool(self, pool_name: str, pool_object: Dict):
        """Create ip pool via Fortigate API

        :param pool_name: Name of ip pool on firewall
        :param pool_object: Dict representation of ip pool
        """
        return self.fortigate.create_firewall_ippool(pool_name, pool_object)

    def update_firewall_ippool(self, pool_name: str, pool_object: Dict):
        """Update ip pool via Fortigate API

        :param pool_name: Name of ip pool on firewall
        :param pool_object: Dict representation of ip pool
        """
        return self.fortigate.update_firewall_ippool(pool_name, pool_object)

    def delete_firewall_ippool(self, pool_name: str):
        """Delete ip pool via Fortigate API

        :param pool_name: Name of ip pool on firewall
        """
        return self.fortigate.delete_firewall_ippool(pool_name)

    def get_firewall_policies(self):
        """Get policies via Fortigate API"""
        return self.fortigate.get_firewall_policy()

    def create_firewall_policy(self, policy_id: int, policy_object: Dict):
        """Create policy via Fortigate API

        :param policy_id: Name of policy on firewall
        :param policy_object: Dict representation of policy
        """
        return self.fortigate.create_firewall_policy(
            str(policy_id), policy_object
        )

    def update_firewall_policy(self, policy_id: int, policy_object: Dict):
        """Update policy via Fortigate API

        :param policy_id: Name of policy on firewall
        :param policy_object: Dict representation of policy
        """
        return self.fortigate.update_firewall_policy(
            str(policy_id), policy_object
        )

    def move_firewall_policy(
        self,
        policy_id: int,
        move_direction: FortiGateApiPolicyDirection,
        move_identifier: int,
    ):
        """Update position of policy via Fortigate API

        :param policy_id: ID of policy on firewall
        :param move_direction: direction after or before another policy
        :param move_identifier: ID of policy on firewall, where policy should move before or after
        """

        return self.fortigate.move_firewall_policy(
            str(policy_id),
            move_direction,
            str(move_identifier),
        )

    def move_firewall_proxy_policy(
        self,
        proxy_policy_id: int,
        move_direction: FortiGateApiPolicyDirection,
        move_identifier: int,
    ):
        """Update position of proxy-policy via Fortigate API

        :param proxy_policy_id: ID of proxy-policy on firewall
        :param move_direction: direction after or before another policy
        :param move_identifier: ID of policy on firewall, where policy schulde move before or after
        """

        return self.fortigate.move_firewall_proxy_policy(
            str(proxy_policy_id),
            move_direction,
            str(move_identifier),
        )

    def delete_firewall_policy(self, policy_id: int):
        """Delete policy via Fortigate API

        :param policy_id: Name of policy on firewall
        """
        return self.fortigate.delete_firewall_policy(str(policy_id))

    def get_firewall_proxy_address(self):
        """Get proxy addresses via Fortigate API"""
        return self.fortigate.get_firewall_proxy_address()

    def create_firewall_proxy_address(
        self, address_name: str, address_object: Dict
    ):
        """Create proxy address via Fortigate API

        :param address_name: Name of proxy address on firewall
        :param address_object: Dict representation of proxy address
        """
        return self.fortigate.create_firewall_proxy_address(
            address_name, address_object
        )

    def update_firewall_proxy_address(
        self, address_name: str, address_object: Dict
    ):
        """Update proxy address via Fortigate API

        :param address_name: Name of proxy address on firewall
        :param address_object: Dict representation of proxy address
        """
        return self.fortigate.update_firewall_proxy_address(
            address_name, address_object
        )

    def delete_firewall_proxy_address(self, address_name: str):
        """Delete proxy address via Fortigate API

        :param address_name: Name of proxy address on firewall
        """
        return self.fortigate.delete_firewall_proxy_address(address_name)

    def get_firewall_proxy_policies(self):
        """Get proxy policies via Fortigate API"""
        return self.fortigate.get_firewall_proxy_policy()

    def get_firewall_proxy_address_group(self):
        """Get proxy address groups via Fortigate API"""
        return self.fortigate.get_firewall_proxy_address_group()

    def create_firewall_proxy_address_group(
        self, group_name: str, group_object: Dict
    ):
        """Create proxy address group via Fortigate API

        :param group_name: Name of address group on firewall
        :param group_object: Dict representation of address group
        """
        return self.fortigate.create_proxy_address_group(
            group_name,
            group_object,
        )

    def update_firewall_proxy_address_group(
        self, group_name: str, group_object: Dict
    ):
        """Update proxy address group via Fortigate API

        :param group_name: Name of address group on firewall
        :param group_object: Dict representation of address group
        """
        return self.fortigate.update_proxy_address_group(
            group_name,
            group_object,
        )

    def delete_firewall_proxy_address_group(self, group_name: str):
        """Delete proxy address group via Fortigate API

        :param group_name: Name of proxy address group on firewall
        """
        return self.fortigate.delete_proxy_address_group(group_name)

    def create_firewall_proxy_policies(
        self, policy_id: int, proxy_policy_object: Dict
    ):
        """Create proxy policy via Fortigate API

        :param policy_id: ID of proxy policy on firewall
        :param proxy_policy_object: Dict representation of proxy policy
        """
        return self.fortigate.create_firewall_proxy_policy(
            str(policy_id), proxy_policy_object
        )

    def update_firewall_proxy_policies(
        self, policy_id: int, proxy_policy_object: Dict
    ):
        """Update policy via Fortigate API

        :param policy_id: ID of proxy policy on firewall
        :param proxy_policy_object: Dict representation of proxy policy
        """
        return self.fortigate.update_firewall_proxy_policy(
            str(policy_id), proxy_policy_object
        )

    def delete_firewall_proxy_policies(self, policy_id: int):
        """Delete proxy policy via Fortigate API

        :param policy_id: ID of proxy policy on firewall
        """
        return self.fortigate.delete_firewall_proxy_policy(str(policy_id))

    def get_firewall_phase1_interface(self):
        """Get phase1 interfaces via Fortigate API"""
        return self.fortigate.get_firewall_phase1_interface()

    def create_firewall_phase1_interface(
        self, name: str, phase1_interface_object: Dict
    ):
        """Create phase1 interface via Fortigate API

        :param name: Name of phase1 interface on firewall
        :param phase1_interface_object: Dict representation of phase1 interface
        """
        return self.fortigate.create_firewall_phase1_interface(
            name, phase1_interface_object
        )

    def update_firewall_phase1_interface(
        self, name: str, phase1_interface_object: Dict
    ):
        """Update phase1 interface via Fortigate API

        :param name: Name of phase1 interface on firewall
        :param phase1_interface_object: Dict representation of phase1 interface
        """
        return self.fortigate.update_firewall_phase1_interface(
            name, phase1_interface_object
        )

    def delete_firewall_phase1_interface(self, name: str):
        """Delete phase1 interfacevia Fortigate API

        :param name: Name of phase1 interface on firewall
        """
        return self.fortigate.delete_firewall_phase1_interface(name)

    def get_firewall_phase2_interface(self):
        """Get phase2 interfaces via Fortigate API"""
        return self.fortigate.get_firewall_phase2_interface()

    def create_firewall_phase2_interface(
        self, name: str, phase2_interface_object: Dict
    ):
        """Create phase2 interface via Fortigate API

        :param name: Name of phase2 interface on firewall
        :param phase2_interface_object: Dict representation of phase2 interface
        """
        return self.fortigate.create_firewall_phase2_interface(
            name, phase2_interface_object
        )

    def update_firewall_phase2_interface(
        self, name: str, phase2_interface_object: Dict
    ):
        """Update phase2 interface via Fortigate API

        :param name: Name of phase2 interface on firewall
        :param phase2_interface_object: Dict representation of phase2 interface
        """
        return self.fortigate.update_firewall_phase2_interface(
            name, phase2_interface_object
        )

    def delete_firewall_phase2_interface(self, name: str):
        """Delete phase2 interfacevia Fortigate API

        :param name: Name of phase2 interface on firewall
        """
        return self.fortigate.delete_firewall_phase2_interface(name)


class FortiGateQueryType(Enum):
    GET = 1
    PUT = 2
    POST = 3
    DELETE = 4


class FortiGateOperation:
    def __init__(
        self,
        url: str,
        query_type: FortiGateQueryType,
        identifier: str,
        data: dict | None = None,
        params: dict | None = None,
    ):
        self.url: str = url
        if params:
            self.url = f"{self.url}?{urllib.parse.urlencode(params)}"

        self.query_type: FortiGateQueryType = query_type
        self.identifier: str = identifier
        self.data: dict = data or dict()

    def __repr__(self):
        return (
            f"{self.query_type.name} {self.url} {self.identifier} {self.data}"
        )


class FortiGateApi:
    ENDPOINT_FIREWALL_INTERFACE = "api/v2/cmdb/system/interface/"
    ENDPOINT_FIREWALL_PHASE1_INTERFACE = (
        "api/v2/cmdb/vpn.ipsec/phase1-interface/"
    )
    ENDPOINT_FIREWALL_PHASE2_INTERFACE = (
        "api/v2/cmdb/vpn.ipsec/phase2-interface/"
    )
    ENDPOINT_FIREWALL_IPPOOL = "api/v2/cmdb/firewall/ippool/"
    ENDPOINT_FIREWALL_VIP = "api/v2/cmdb/firewall/vip/"
    ENDPOINT_FIREWALL_VIP_GROUP = "api/v2/cmdb/firewall/vipgrp/"
    ENDPOINT_FIREWALL_POLICY = "api/v2/cmdb/firewall/policy/"
    ENDPOINT_FIREWALL_ADDRESS = "api/v2/cmdb/firewall/address/"
    ENDPOINT_FIREWALL_ADDRESS_GROUP = "api/v2/cmdb/firewall/addrgrp/"
    ENDPOINT_FIREWALL_SERVICE = "api/v2/cmdb/firewall.service/custom/"
    ENDPOINT_FIREWALL_SERVICE_GROUP = "api/v2/cmdb/firewall.service/group/"
    ENDPOINT_IPSEC_VPN = "api/v2/cmdb/vpn.ipsec/"
    ENDPOINT_IPSEC_VPN_MONITOR = "api/v2/monitor/vpn/ipsec/"
    ENDPOINT_STATIC_ROUTE = "api/v2/cmdb/router/static/"
    ENDPOINT_FIREWALL_PROXY_ADDRESS = "api/v2/cmdb/firewall/proxy-address/"
    ENDPOINT_FIREWALL_PROXY_ADDRESS_GROUP = (
        "api/v2/cmdb/firewall/proxy-addrgrp/"
    )
    ENDPOINT_FIREWALL_PROXY_POLICY = "api/v2/cmdb/firewall/proxy-policy/"

    def __init__(
        self,
        ipaddr: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        access_token: Optional[str] = None,
        timeout: int = 10,
        vdom: str = "root",
        port: str = "443",
        read_only: bool = False,
    ):
        if not any([username, password, access_token]):
            raise InternalErrorException("No Login Method given!")

        self.ipaddr: str = ipaddr
        self.username: str = username
        self.password: str = password
        self.access_token: str = access_token
        self.port: str = port
        self.urlbase: str = f"https://{self.ipaddr}:{self.port}/"
        self.timeout: int = timeout
        self.vdom: str = vdom
        self.read_only: bool = read_only

        self.client: httpx.Client = None
        self.operations: List[FortiGateOperation] = []

    def login(self):
        """Login via Username or Access Token.
        Get CSRF Token and use it for active session.
        """
        self.client = httpx.Client(verify=False, timeout=self.timeout)

        # access token? then use is
        if self.access_token:
            self.login_with_access_token()
        else:  # no token? then use username and password login
            self.login_with_username()

        # Check whether login was successful
        login_check = self.client.get(self.urlbase + "api/v2/cmdb/system/vdom")
        self.check_response_code(login_check)

    def login_with_username(self):
        """Login via Fortigate API.
        Get CSRF Token and use it for active session.
        """
        url = self.urlbase + "logincheck"
        response = self.client.post(
            url,
            data={"username": self.username, "secretkey": self.password},
        )

        self.check_response_code(response)

        # Get CSRF token from cookies, add to headers
        if "ccsrftoken" in response.cookies:
            csrftoken = response.cookies["ccsrftoken"][1:-1]
            self.client.headers.update({"X-CSRFTOKEN": csrftoken})

    def login_with_access_token(self):
        self.client.headers.update(
            {"Authorization": f"Bearer {self.access_token}"}
        )

    def logout(self):
        """Logout via Fortigate API.
        Logout the active session.
        """

        if not self.access_token:
            url = self.urlbase + "logout"
            self.client.get(url)

    def does_exist(self, object_url: str, params: dict | None = None) -> bool:
        params = params or dict()
        params.update(vdom=self.vdom)

        response = self.client.get(
            object_url,
            params=params,
        )
        if response.status_code == 200:
            return True
        return False

    def get(self, url, params: dict | None = None):
        params = params or dict()
        params.update(vdom=self.vdom)

        return self.client.get(
            url,
            params=params,
        )

    def put(self, url, data: dict | None, params: dict | None = None):
        params = params or dict()
        params.update(vdom=self.vdom)

        return self.client.put(
            url,
            json=data,
            params=params,
        )

    def post(self, url, data: dict, params: dict | None = None):
        params = params or dict()
        params.update(vdom=self.vdom)

        return self.client.post(
            url,
            json=data,
            params=params,
        )

    def delete(self, url, params: dict | None = None):
        params = params or dict()
        params.update(vdom=self.vdom)

        return self.client.delete(
            url,
            params=params,
        )

    def check_response_code(self, response: httpx.Response):
        if response.status_code == 200:
            return
        elif response.status_code >= 300 and response.status_code < 400:
            return
        elif response.status_code == 400:
            raise BadRequestExeption(response)
        elif response.status_code == 401:
            raise NotAuthorizedException(response)
        elif response.status_code == 403:
            raise ForbiddenException(response)
        elif response.status_code == 404:
            raise ResourceNotFoundException(response)
        elif response.status_code == 405:
            raise MethodNotAllowedException(response)
        elif response.status_code == 413:
            raise RequestEntityTooLargeException(response)
        elif response.status_code == 424:
            raise FailedDependencyException(response)
        elif response.status_code == 429:
            raise TooManyRequestsException(response)
        elif response.status_code == 500:
            raise InternalErrorException(response)
        else:
            raise APIException(response)

    def query_api_get(
        self, uri: str, specific=False, filters=False
    ) -> Union[dict, int]:
        api_url = self.urlbase + uri
        params = dict()

        if specific:
            api_url += specific
        if filters:
            params.update(filter=filters)

        self.operations.append(
            FortiGateOperation(api_url, FortiGateQueryType.GET, "", "")
        )

        result = self.get(api_url, params)
        self.check_response_code(result)
        return result.json()["results"]

    def query_api_create(
        self, uri: str, identifier: str, data: Dict
    ) -> Union[dict, int]:
        api_url = self.urlbase + uri
        if self.does_exist(api_url + identifier):
            raise ObjectAlreadyExitsException(
                f"object {identifier} " f"already exists"
            )

        self.operations.append(
            FortiGateOperation(
                api_url, FortiGateQueryType.POST, identifier, data
            )
        )
        if self.read_only:
            return {}
        result = self.post(api_url, data)
        self.check_response_code(result)
        return result.json()

    def query_api_update(
        self,
        uri: str,
        identifier: str,
        data: Dict,
    ) -> Union[dict, int]:
        api_url = self.urlbase + uri
        self.operations.append(
            FortiGateOperation(
                api_url, FortiGateQueryType.PUT, identifier, data
            )
        )

        if self.read_only:
            return {}

        result = self.put(f"{api_url}/{identifier}", data)
        self.check_response_code(result)
        return result.json()

    def query_api_move(
        self,
        uri: str,
        identifier: str,
        move_direction: FortiGateApiPolicyDirection = None,
        move_identifier: str = None,
    ) -> Union[dict, int]:
        api_url = self.urlbase + uri
        move_url = f"{api_url}/{identifier}"
        params = {
            "action": "move",
            move_direction.value: move_identifier,
        }
        self.operations.append(
            FortiGateOperation(
                move_url,
                FortiGateQueryType.PUT,
                identifier,
                data=None,
                params=params,
            )
        )

        if self.read_only:
            return {}

        result = self.put(
            move_url,
            data=None,
            params=params,
        )
        self.check_response_code(result)
        return result.json()

    def query_api_delete(self, uri: str, identifier: str) -> Union[dict, int]:
        api_url = self.urlbase + uri

        self.operations.append(
            FortiGateOperation(
                api_url, FortiGateQueryType.DELETE, identifier, ""
            )
        )

        if self.read_only:
            return {}

        result = self.delete(f"{api_url}/{identifier}")
        self.check_response_code(result)
        return result.json()

    # Firewall Route Methods
    def get_firewall_route_static(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_STATIC_ROUTE, specific, filters
        )

    def create_firewall_route_static(self, seq_num: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_STATIC_ROUTE, seq_num, data
        )

    def update_firewall_route_static(self, seq_num: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_STATIC_ROUTE, seq_num, data
        )

    def delete_firewall_route_static(self, seq_num: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_STATIC_ROUTE, seq_num
        )

    # Firewall Interface Methods
    def get_firewall_interface(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_INTERFACE, specific, filters
        )

    def get_ipsec_vpn(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_IPSEC_VPN, specific, filters
        )

    # Firewall Address Methods
    def get_ipsec_vpn_monitor(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_IPSEC_VPN_MONITOR, specific, filters
        )

    def get_firewall_vip(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_VIP, specific, filters
        )

    def get_firewall_vip_group(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_VIP_GROUP, specific, filters
        )

    def get_firewall_ippool(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_IPPOOL, specific, filters
        )

    def get_firewall_policy(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_POLICY, specific, filters
        )

    def get_firewall_address(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_ADDRESS, specific, filters
        )

    def get_address_group(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_ADDRESS_GROUP, specific, filters
        )

    def get_firewall_service(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_SERVICE, specific, filters
        )

    def get_service_group(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_SERVICE_GROUP, specific, filters
        )

    def get_firewall_proxy_address(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_ADDRESS, specific, filters
        )

    def get_firewall_proxy_address_group(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_ADDRESS_GROUP,
            specific,
            filters,
        )

    def get_firewall_proxy_policy(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_POLICY, specific, filters
        )

    def create_firewall_ippool(self, pool_name: str, data: Dict):
        """Query Fortigate API to create ip pool.

        :param pool_name: Name of ip pool
        :param data: JSON object representation
        """
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_IPPOOL, pool_name, data
        )

    def update_firewall_ippool(self, pool_name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_IPPOOL, pool_name, data
        )

    def delete_firewall_ippool(self, pool_name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_IPPOOL, pool_name
        )

    def create_firewall_vip(self, vip_name: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_VIP, vip_name, data
        )

    def update_firewall_vip(self, vip_name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_VIP, vip_name, data
        )

    def delete_firewall_vip(self, vip_name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_VIP, vip_name
        )

    def create_firewall_vip_group(self, vip_group_name: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_VIP_GROUP, vip_group_name, data
        )

    def update_firewall_vip_group(self, vip_group_name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_VIP_GROUP, vip_group_name, data
        )

    def delete_firewall_vip_group(self, vip_group_name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_VIP_GROUP, vip_group_name
        )

    def create_firewall_address(self, address_name: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_ADDRESS, address_name, data
        )

    def update_firewall_address(self, address_name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_ADDRESS, address_name, data
        )

    def delete_firewall_address(self, address_name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_ADDRESS, address_name
        )

    def create_address_group(self, group_name: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_ADDRESS_GROUP, group_name, data
        )

    def update_address_group(self, group_name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_ADDRESS_GROUP, group_name, data
        )

    def delete_address_group(self, group_name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_ADDRESS_GROUP, group_name
        )

    def create_firewall_service(self, service_name: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_SERVICE, service_name, data
        )

    def update_firewall_service(self, service_name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_SERVICE, service_name, data
        )

    def delete_firewall_service(self, service_name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_SERVICE, service_name
        )

    def create_service_group(self, group_name: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_SERVICE_GROUP, group_name, data
        )

    def update_service_group(self, group_name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_SERVICE_GROUP, group_name, data
        )

    def delete_service_group(self, group_name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_SERVICE_GROUP, group_name
        )

    def create_firewall_policy(self, policy_id: str, data: Dict):
        ret = self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_POLICY, policy_id, data
        )
        if self.read_only:
            return {"mkey": "99999999"}

        return ret

    def update_firewall_policy(self, policy_id: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_POLICY, policy_id, data
        )

    def move_firewall_policy(
        self,
        policy_id: str,
        move_direction: FortiGateApiPolicyDirection,
        move_identifier: str,
    ):
        return self.query_api_move(
            FortiGateApi.ENDPOINT_FIREWALL_POLICY,
            policy_id,
            move_direction,
            move_identifier,
        )

    def move_firewall_proxy_policy(
        self,
        proxy_policy_id: str,
        move_direction: FortiGateApiPolicyDirection,
        move_identifier: str,
    ):
        return self.query_api_move(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_POLICY,
            proxy_policy_id,
            move_direction,
            move_identifier,
        )

    def delete_firewall_policy(self, policy_id: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_POLICY, policy_id
        )

    def create_firewall_proxy_address(self, address_name: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_ADDRESS, address_name, data
        )

    def update_firewall_proxy_address(self, address_name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_ADDRESS,
            urllib.parse.quote_plus(address_name),
            data,
        )

    def delete_firewall_proxy_address(self, address_name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_ADDRESS,
            urllib.parse.quote_plus(address_name),
        )

    def create_firewall_proxy_policy(self, policy_id: str, data: Dict):
        ret = self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_POLICY, policy_id, data
        )
        if self.read_only:
            return {"mkey": "99999999"}

        return ret

    def update_firewall_proxy_policy(self, policy_id: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_POLICY, policy_id, data
        )

    def delete_firewall_proxy_policy(self, policy_id: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_POLICY, policy_id
        )

    def create_proxy_address_group(self, group_name: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_ADDRESS_GROUP,
            group_name,
            data,
        )

    def update_proxy_address_group(self, group_name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_ADDRESS_GROUP,
            group_name,
            data,
        )

    def delete_proxy_address_group(self, group_name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_PROXY_ADDRESS_GROUP, group_name
        )

    def get_write_operations(self):
        filtered: List[FortiGateOperation] = []
        for operation in self.operations:
            if operation.query_type != FortiGateQueryType.GET:
                filtered.append(operation)

        return filtered

    def get_firewall_phase1_interface(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_PHASE1_INTERFACE, specific, filters
        )

    def create_firewall_phase1_interface(self, name: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_PHASE1_INTERFACE, name, data
        )

    def update_firewall_phase1_interface(self, name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_PHASE1_INTERFACE, name, data
        )

    def delete_firewall_phase1_interface(self, name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_PHASE1_INTERFACE, name
        )

    def get_firewall_phase2_interface(self, specific=False, filters=False):
        return self.query_api_get(
            FortiGateApi.ENDPOINT_FIREWALL_PHASE2_INTERFACE, specific, filters
        )

    def create_firewall_phase2_interface(self, name: str, data: Dict):
        return self.query_api_create(
            FortiGateApi.ENDPOINT_FIREWALL_PHASE2_INTERFACE, name, data
        )

    def update_firewall_phase2_interface(self, name: str, data: Dict):
        return self.query_api_update(
            FortiGateApi.ENDPOINT_FIREWALL_PHASE2_INTERFACE, name, data
        )

    def delete_firewall_phase2_interface(self, name: str):
        return self.query_api_delete(
            FortiGateApi.ENDPOINT_FIREWALL_PHASE2_INTERFACE, name
        )
